@@ -1,6 +1,8 @@
 /*
- * Copyright (C) 2007-2008 B.A.T.M.A.N. contributors:
+ * Copyright (C) 2007-2009 B.A.T.M.A.N. contributors:
+ *
  * Marek Lindner, Simon Wunderlich
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of version 2 of the GNU General Public
  * License as published by the Free Software Foundation.
@@ -33,15 +35,54 @@
 #include "hash.h"
 #include "ring_buffer.h"
 #include "vis.h"
+#include "aggregation.h"
+#include "compat.h"
 
 
 
 DECLARE_WAIT_QUEUE_HEAD(thread_wait);
+static DECLARE_DELAYED_WORK(purge_orig_wq, purge_orig);
 
 static atomic_t data_ready_cond;
 atomic_t exit_cond;
 
+static void start_purge_timer(void)
+{
+	queue_delayed_work(bat_event_workqueue, &purge_orig_wq, 1 * HZ);
+}
 
+int originator_init(void)
+{
+	if (orig_hash)
+		return 1;
+
+	spin_lock(&orig_hash_lock);
+	orig_hash = hash_new(128, compare_orig, choose_orig);
+
+	if (!orig_hash)
+		goto err;
+
+	spin_unlock(&orig_hash_lock);
+	start_purge_timer();
+	return 1;
+
+err:
+	spin_unlock(&orig_hash_lock);
+	return 0;
+}
+
+void originator_free(void)
+{
+	if (!orig_hash)
+		return;
+
+	cancel_delayed_work_sync(&purge_orig_wq);
+
+	spin_lock(&orig_hash_lock);
+	hash_delete(orig_hash, free_orig_node);
+	orig_hash = NULL;
+	spin_unlock(&orig_hash_lock);
+}
 
 static struct neigh_node *create_neighbor(struct orig_node *orig_node, struct orig_node *orig_neigh_node, uint8_t *neigh, struct batman_if *if_incoming)
 {
@@ -49,7 +90,7 @@ static struct neigh_node *create_neighbor(struct orig_node *orig_node, struct or
 
 	debug_log(LOG_TYPE_BATMAN, "Creating new last-hop neighbour of originator\n");
 
-	neigh_node = kmalloc(sizeof(struct neigh_node), GFP_KERNEL);
+	neigh_node = kmalloc(sizeof(struct neigh_node), GFP_ATOMIC);
 	memset(neigh_node, 0, sizeof(struct neigh_node));
 	INIT_LIST_HEAD(&neigh_node->list);
 
@@ -75,7 +116,7 @@ void free_orig_node(void *data)
 		kfree(neigh_node);
 	}
 
-	hna_global_del_orig(orig_node, "originator timeouted");
+	hna_global_del_orig(orig_node, "originator timed out");
 
 	kfree(orig_node->bcast_own);
 	kfree(orig_node->bcast_own_sum);
@@ -97,7 +138,7 @@ static struct orig_node *get_orig_node(uint8_t *addr)
 	addr_to_string(orig_str, addr);
 	debug_log(LOG_TYPE_BATMAN, "Creating new originator: %s \n", orig_str);
 
-	orig_node = kmalloc(sizeof(struct orig_node), GFP_KERNEL);
+	orig_node = kmalloc(sizeof(struct orig_node), GFP_ATOMIC);
 	memset(orig_node, 0, sizeof(struct orig_node));
 	INIT_LIST_HEAD(&orig_node->neigh_list);
 
@@ -106,10 +147,10 @@ static struct orig_node *get_orig_node(uint8_t *addr)
 	orig_node->batman_if = NULL;
 	orig_node->hna_buff = NULL;
 
-	orig_node->bcast_own = kmalloc(num_ifs * sizeof(TYPE_OF_WORD) * NUM_WORDS, GFP_KERNEL);
+	orig_node->bcast_own = kmalloc(num_ifs * sizeof(TYPE_OF_WORD) * NUM_WORDS, GFP_ATOMIC);
 	memset(orig_node->bcast_own, 0, num_ifs * sizeof(TYPE_OF_WORD) * NUM_WORDS);
 
-	orig_node->bcast_own_sum = kmalloc(num_ifs * sizeof(uint8_t), GFP_KERNEL);
+	orig_node->bcast_own_sum = kmalloc(num_ifs * sizeof(uint8_t), GFP_ATOMIC);
 	memset(orig_node->bcast_own_sum, 0, num_ifs * sizeof(uint8_t));
 
 	hash_add(orig_hash, orig_node);
@@ -157,7 +198,7 @@ static void update_routes(struct orig_node *orig_node, struct neigh_node *neigh_
 		if ((orig_node->router != NULL) && (neigh_node == NULL)) {
 
 			debug_log(LOG_TYPE_ROUTES, "Deleting route towards: %s\n", orig_str);
-			hna_global_del_orig(orig_node, "originator timeouted");
+			hna_global_del_orig(orig_node, "originator timed out");
 
 		/* route added */
 		} else if ((orig_node->router == NULL) && (neigh_node != NULL)) {
@@ -185,7 +226,7 @@ static void update_routes(struct orig_node *orig_node, struct neigh_node *neigh_
 	/* may be just HNA changed */
 	} else {
 
-		if ((hna_buff_len != orig_node->hna_buff_len) || ((hna_buff_len > 0 ) && (orig_node->hna_buff_len > 0) && (memcmp(orig_node->hna_buff, hna_buff, hna_buff_len) != 0))) {
+		if ((hna_buff_len != orig_node->hna_buff_len) || ((hna_buff_len > 0) && (orig_node->hna_buff_len > 0) && (memcmp(orig_node->hna_buff, hna_buff, hna_buff_len) != 0))) {
 
 			if (orig_node->hna_buff_len > 0)
 				hna_global_del_orig(orig_node, "originator changed hna");
@@ -226,7 +267,7 @@ static int isBidirectionalNeigh(struct orig_node *orig_node, struct orig_node *o
 				neigh_node = tmp_neigh_node;
 		}
 
-		if ( neigh_node == NULL )
+		if (neigh_node == NULL)
 			neigh_node = create_neighbor(orig_neigh_node, orig_neigh_node, orig_neigh_node->orig, if_incoming);
 	}
 
@@ -246,7 +287,7 @@ static int isBidirectionalNeigh(struct orig_node *orig_node, struct orig_node *o
 	/*
 	 * 1 - ((1-x) ** 3), normalized to TQ_MAX_VALUE
 	 * this does affect the nearly-symmetric links only a little,
-	 * but punishes asymetric links more.
+	 * but punishes asymmetric links more.
 	 * this will give a value between 0 and TQ_MAX_VALUE
 	 */
 	orig_neigh_node->tq_asym_penalty = TQ_MAX_VALUE - (TQ_MAX_VALUE *
@@ -257,7 +298,7 @@ static int isBidirectionalNeigh(struct orig_node *orig_node, struct orig_node *o
 
 	batman_packet->tq = ((batman_packet->tq * orig_neigh_node->tq_own * orig_neigh_node->tq_asym_penalty) / (TQ_MAX_VALUE *  TQ_MAX_VALUE));
 
-	debug_log(LOG_TYPE_BATMAN, "bidirectional: orig = %-15s neigh = %-15s => own_bcast = %2i, real recv = %2i, local tq: %3i, asym_penality: %3i, total tq: %3i \n",
+	debug_log(LOG_TYPE_BATMAN, "bidirectional: orig = %-15s neigh = %-15s => own_bcast = %2i, real recv = %2i, local tq: %3i, asym_penalty: %3i, total tq: %3i \n",
 		  orig_str, neigh_str, total_count, neigh_node->real_packet_count, orig_neigh_node->tq_own, orig_neigh_node->tq_asym_penalty, batman_packet->tq);
 
 	/* if link has the minimum required transmission quality consider it bidirectional */
@@ -269,31 +310,22 @@ static int isBidirectionalNeigh(struct orig_node *orig_node, struct orig_node *o
 
 static void update_orig(struct orig_node *orig_node, struct ethhdr *ethhdr, struct batman_packet *batman_packet, struct batman_if *if_incoming, unsigned char *hna_buff, int hna_buff_len, char is_duplicate)
 {
-	struct neigh_node *neigh_node = NULL, *tmp_neigh_node = NULL, *best_neigh_node = NULL;
-	unsigned char max_tq = 0, max_bcast_own = 0;
+	struct neigh_node *neigh_node = NULL, *tmp_neigh_node = NULL;
 	int tmp_hna_buff_len;
 
 	debug_log(LOG_TYPE_BATMAN, "update_originator(): Searching and updating originator entry of received packet \n");
 
 	list_for_each_entry(tmp_neigh_node, &orig_node->neigh_list, list) {
-
 		if (compare_orig(tmp_neigh_node->addr, ethhdr->h_source) && (tmp_neigh_node->if_incoming == if_incoming)) {
 			neigh_node = tmp_neigh_node;
-		} else {
-
-			if (!is_duplicate) {
-				ring_buffer_set(tmp_neigh_node->tq_recv, &tmp_neigh_node->tq_index, 0);
-				tmp_neigh_node->tq_avg = ring_buffer_avg(tmp_neigh_node->tq_recv);
-			}
-
-			/* if we got have a better tq value via this neighbour or same tq value if it is currently our best neighbour (to avoid route flipping) */
-			if ((tmp_neigh_node->tq_avg > max_tq) || ((tmp_neigh_node->tq_avg == max_tq) && (tmp_neigh_node->orig_node->bcast_own_sum[if_incoming->if_num] > max_bcast_own)) || ((orig_node->router == tmp_neigh_node) && (tmp_neigh_node->tq_avg == max_tq))) {
-
-				max_tq = tmp_neigh_node->tq_avg;
-				max_bcast_own = tmp_neigh_node->orig_node->bcast_own_sum[if_incoming->if_num];
-				best_neigh_node = tmp_neigh_node;
-			}
+			continue;
 		}
+
+		if (is_duplicate)
+			continue;
+
+		ring_buffer_set(tmp_neigh_node->tq_recv, &tmp_neigh_node->tq_index, 0);
+		tmp_neigh_node->tq_avg = ring_buffer_avg(tmp_neigh_node->tq_recv);
 	}
 
 	if (neigh_node == NULL)
@@ -312,17 +344,30 @@ static void update_orig(struct orig_node *orig_node, struct ethhdr *ethhdr, stru
 		neigh_node->last_ttl = batman_packet->ttl;
 	}
 
-	if ((neigh_node->tq_avg > max_tq) || ((neigh_node->tq_avg == max_tq) && (neigh_node->orig_node->bcast_own_sum[if_incoming->if_num] > max_bcast_own)) || ((orig_node->router == neigh_node) && (neigh_node->tq_avg == max_tq))) {
-
-		max_tq = neigh_node->tq_avg;
-		max_bcast_own = neigh_node->orig_node->bcast_own_sum[if_incoming->if_num];
-		best_neigh_node = neigh_node;
-
-	}
-
 	tmp_hna_buff_len = (hna_buff_len > batman_packet->num_hna * ETH_ALEN ? batman_packet->num_hna * ETH_ALEN : hna_buff_len);
 
-	update_routes(orig_node, best_neigh_node, hna_buff, tmp_hna_buff_len);
+	/* if this neighbor already is our next hop there is nothing to change */
+	if (orig_node->router == neigh_node)
+		goto update_hna;
+
+	/* if this neighbor does not offer a better TQ we won't consider it */
+	if ((orig_node->router) &&
+	    (orig_node->router->tq_avg > neigh_node->tq_avg))
+		goto update_hna;
+
+	/* if the TQ is the same and the link not more symetric we won't consider it either */
+	if ((orig_node->router) &&
+	     ((neigh_node->tq_avg == orig_node->router->tq_avg) &&
+	     (orig_node->router->orig_node->bcast_own_sum[if_incoming->if_num] >
+	      neigh_node->orig_node->bcast_own_sum[if_incoming->if_num])))
+		goto update_hna;
+
+	update_routes(orig_node, neigh_node, hna_buff, tmp_hna_buff_len);
+	return;
+
+update_hna:
+	update_routes(orig_node, orig_node->router, hna_buff, tmp_hna_buff_len);
+	return;
 }
 
 static char count_real_packets(struct ethhdr *ethhdr, struct batman_packet *batman_packet, struct batman_if *if_incoming)
@@ -333,7 +378,7 @@ static char count_real_packets(struct ethhdr *ethhdr, struct batman_packet *batm
 
 
 	orig_node = get_orig_node(batman_packet->orig);
-	if (orig_node == NULL) 
+	if (orig_node == NULL)
 		return 0;
 
 
@@ -358,30 +403,43 @@ static char count_real_packets(struct ethhdr *ethhdr, struct batman_packet *batm
 	return is_duplicate;
 }
 
-static void receive_bat_packet(struct ethhdr *ethhdr, struct batman_packet *batman_packet, unsigned char *hna_buff, int hna_buff_len, struct batman_if *if_incoming)
+void receive_bat_packet(struct ethhdr *ethhdr, struct batman_packet *batman_packet, unsigned char *hna_buff, int hna_buff_len, struct batman_if *if_incoming)
 {
 	struct batman_if *batman_if;
 	struct orig_node *orig_neigh_node, *orig_node;
-	char orig_str[ETH_STR_LEN], old_orig_str[ETH_STR_LEN], neigh_str[ETH_STR_LEN];
+	char orig_str[ETH_STR_LEN], prev_sender_str[ETH_STR_LEN], neigh_str[ETH_STR_LEN];
 	char has_directlink_flag;
 	char is_my_addr = 0, is_my_orig = 0, is_my_oldorig = 0, is_broadcast = 0, is_bidirectional, is_single_hop_neigh, is_duplicate;
 	unsigned short if_incoming_seqno;
 
-	/* could be changed by send_own_packet() */
+	/* Silently drop when the batman packet is actually not a correct packet.
+	 *
+	 * This might happen if a packet is padded (e.g. Ethernet has a
+	 * minimum frame length of 64 byte) and the aggregation interprets
+	 * it as an additional length.
+	 *
+	 * TODO: A more sane solution would be to have a bit in the batman_packet
+	 * to detect whether the packet is the last packet in an aggregation.
+	 * Here we expect that the padding is always zero (or not 0x01)
+	 */
+	if (batman_packet->packet_type != BAT_PACKET)
+		return;
+
+	/* could be changed by schedule_own_packet() */
 	if_incoming_seqno = atomic_read(&if_incoming->seqno);
 
 	addr_to_string(orig_str, batman_packet->orig);
-	addr_to_string(old_orig_str, batman_packet->old_orig);
+	addr_to_string(prev_sender_str, batman_packet->prev_sender);
 	addr_to_string(neigh_str, ethhdr->h_source);
 
 	has_directlink_flag = (batman_packet->flags & DIRECTLINK ? 1 : 0);
 
 	is_single_hop_neigh = (compare_orig(ethhdr->h_source, batman_packet->orig) ? 1 : 0);
 
-	debug_log(LOG_TYPE_BATMAN, "Received BATMAN packet via NB: %s, IF: %s [%s] (from OG: %s, via old OG: %s, seqno %d, tq %d, TTL %d, V %d, IDF %d) \n", neigh_str, if_incoming->dev, if_incoming->addr_str, orig_str, old_orig_str, batman_packet->seqno, batman_packet->tq, batman_packet->ttl, batman_packet->version, has_directlink_flag);
+	debug_log(LOG_TYPE_BATMAN, "Received BATMAN packet via NB: %s, IF: %s [%s] (from OG: %s, via prev OG: %s, seqno %d, tq %d, TTL %d, V %d, IDF %d) \n", neigh_str, if_incoming->dev, if_incoming->addr_str, orig_str, prev_sender_str, batman_packet->seqno, batman_packet->tq, batman_packet->ttl, batman_packet->version, has_directlink_flag);
 
 	list_for_each_entry_rcu(batman_if, &if_list, list) {
-		if (batman_if->if_active != IF_ACTIVE) 
+		if (batman_if->if_active != IF_ACTIVE)
 			continue;
 
 		if (compare_orig(ethhdr->h_source, batman_if->net_dev->dev_addr))
@@ -390,7 +448,7 @@ static void receive_bat_packet(struct ethhdr *ethhdr, struct batman_packet *batm
 		if (compare_orig(batman_packet->orig, batman_if->net_dev->dev_addr))
 			is_my_orig = 1;
 
-		if (compare_orig(batman_packet->old_orig, batman_if->net_dev->dev_addr))
+		if (compare_orig(batman_packet->prev_sender, batman_if->net_dev->dev_addr))
 			is_my_oldorig = 1;
 
 		if (compare_orig(ethhdr->h_source, broadcastAddr))
@@ -398,104 +456,104 @@ static void receive_bat_packet(struct ethhdr *ethhdr, struct batman_packet *batm
 	}
 
 	if (batman_packet->version != COMPAT_VERSION) {
-
 		debug_log(LOG_TYPE_BATMAN, "Drop packet: incompatible batman version (%i) \n", batman_packet->version);
+		return;
+	}
 
-	} else if (is_my_addr) {
-
+	if (is_my_addr) {
 		debug_log(LOG_TYPE_BATMAN, "Drop packet: received my own broadcast (sender: %s) \n", neigh_str);
+		return;
+	}
 
-	} else if (is_broadcast) {
-
+	if (is_broadcast) {
 		debug_log(LOG_TYPE_BATMAN, "Drop packet: ignoring all packets with broadcast source addr (sender: %s) \n", neigh_str);
+		return;
+	}
 
-	} else if (is_my_orig) {
-
+	if (is_my_orig) {
 		orig_neigh_node = get_orig_node(ethhdr->h_source);
 
 		/* neighbour has to indicate direct link and it has to come via the corresponding interface */
 		/* if received seqno equals last send seqno save new seqno for bidirectional check */
-		if (has_directlink_flag && (batman_packet->seqno - if_incoming_seqno + 1 == 0)) {
+		if (has_directlink_flag && compare_orig(if_incoming->net_dev->dev_addr, batman_packet->orig) &&
+				  (batman_packet->seqno - if_incoming_seqno + 2 == 0)) {
 			bit_mark((TYPE_OF_WORD *)&(orig_neigh_node->bcast_own[if_incoming->if_num * NUM_WORDS]), 0);
 			orig_neigh_node->bcast_own_sum[if_incoming->if_num] = bit_packet_count((TYPE_OF_WORD *)&(orig_neigh_node->bcast_own[if_incoming->if_num * NUM_WORDS]));
 		}
 
 		debug_log(LOG_TYPE_BATMAN, "Drop packet: originator packet from myself (via neighbour) \n");
+		return;
+	}
 
-	} else if (batman_packet->tq == 0) {
-
+	if (batman_packet->tq == 0) {
 		count_real_packets(ethhdr, batman_packet, if_incoming);
 
 		debug_log(LOG_TYPE_BATMAN, "Drop packet: originator packet with tq equal 0 \n");
-
-	} else if (is_my_oldorig) {
-
-		debug_log(LOG_TYPE_BATMAN, "Drop packet: ignoring all rebroadcast echos (sender: %s) \n", neigh_str);
-
-	} else {
-
-		is_duplicate = count_real_packets(ethhdr, batman_packet, if_incoming);
-
-		orig_node = get_orig_node(batman_packet->orig);
-		if (orig_node == NULL) 
-			return;
-
-
-		/* if sender is a direct neighbor the sender mac equals originator mac */
-		orig_neigh_node = (is_single_hop_neigh ? orig_node : get_orig_node(ethhdr->h_source));
-		if (orig_neigh_node == NULL) 
-			return;
-
-
-		/* drop packet if sender is not a direct neighbor and if we no route towards it */
-		if (!is_single_hop_neigh && (orig_neigh_node->router == NULL)) {
-
-			debug_log(LOG_TYPE_BATMAN, "Drop packet: OGM via unkown neighbor! \n");
-
-		} else {
-
-			is_bidirectional = isBidirectionalNeigh(orig_node, orig_neigh_node, batman_packet, if_incoming);
-
-			/* update ranking if it is not a duplicate or has the same seqno and similar ttl as the non-duplicate */
-			if (is_bidirectional && (!is_duplicate || ((orig_node->last_real_seqno == batman_packet->seqno) && (orig_node->last_ttl - 3 <= batman_packet->ttl))))
-				update_orig(orig_node, ethhdr, batman_packet, if_incoming, hna_buff, hna_buff_len, is_duplicate);
-
-			/* is single hop (direct) neighbour */
-			if (is_single_hop_neigh) {
-
-				/* mark direct link on incoming interface */
-				send_forward_packet(orig_node, ethhdr, batman_packet, 1, hna_buff, hna_buff_len, if_incoming);
-
-				debug_log(LOG_TYPE_BATMAN, "Forwarding packet: rebroadcast neighbour packet with direct link flag \n");
-
-			/* multihop originator */
-			} else {
-
-				if (is_bidirectional) {
-
-					if (!is_duplicate) {
-
-						send_forward_packet(orig_node, ethhdr, batman_packet, 0, hna_buff, hna_buff_len, if_incoming);
-
-						debug_log(LOG_TYPE_BATMAN, "Forwarding packet: rebroadcast originator packet \n");
-
-					} else {
-
-						debug_log(LOG_TYPE_BATMAN, "Drop packet: duplicate packet received\n");
-
-					}
-
-				} else {
-
-					debug_log(LOG_TYPE_BATMAN, "Drop packet: not received via bidirectional link\n");
-
-				}
-
-			}
-
-		}
-
+		return;
 	}
+
+	if (is_my_oldorig) {
+		debug_log(LOG_TYPE_BATMAN, "Drop packet: ignoring all rebroadcast echos (sender: %s) \n", neigh_str);
+		return;
+	}
+
+	is_duplicate = count_real_packets(ethhdr, batman_packet, if_incoming);
+
+	orig_node = get_orig_node(batman_packet->orig);
+	if (orig_node == NULL)
+		return;
+
+	/* avoid temporary routing loops */
+	if ((orig_node->router) && (orig_node->router->orig_node->router) &&
+	    (compare_orig(orig_node->router->addr, batman_packet->prev_sender)) &&
+	    !(compare_orig(batman_packet->orig, batman_packet->prev_sender)) &&
+	    (compare_orig(orig_node->router->addr, orig_node->router->orig_node->router->addr))) {
+		debug_log(LOG_TYPE_BATMAN, "Drop packet: ignoring all rebroadcast packets that may make me loop (sender: %s) \n", neigh_str);
+		return;
+	}
+
+	/* if sender is a direct neighbor the sender mac equals originator mac */
+	orig_neigh_node = (is_single_hop_neigh ? orig_node : get_orig_node(ethhdr->h_source));
+	if (orig_neigh_node == NULL)
+		return;
+
+	/* drop packet if sender is not a direct neighbor and if we don't route towards it */
+	if (!is_single_hop_neigh && (orig_neigh_node->router == NULL)) {
+		debug_log(LOG_TYPE_BATMAN, "Drop packet: OGM via unknown neighbor! \n");
+		return;
+	}
+
+	is_bidirectional = isBidirectionalNeigh(orig_node, orig_neigh_node, batman_packet, if_incoming);
+
+	/* update ranking if it is not a duplicate or has the same seqno and similar ttl as the non-duplicate */
+	if (is_bidirectional && (!is_duplicate ||
+		((orig_node->last_real_seqno == batman_packet->seqno) &&
+			(orig_node->last_ttl - 3 <= batman_packet->ttl))))
+		update_orig(orig_node, ethhdr, batman_packet, if_incoming, hna_buff, hna_buff_len, is_duplicate);
+
+	/* is single hop (direct) neighbour */
+	if (is_single_hop_neigh) {
+
+		/* mark direct link on incoming interface */
+		schedule_forward_packet(orig_node, ethhdr, batman_packet, 1, hna_buff_len, if_incoming);
+
+		debug_log(LOG_TYPE_BATMAN, "Forwarding packet: rebroadcast neighbour packet with direct link flag \n");
+		return;
+	}
+
+	/* multihop originator */
+	if (!is_bidirectional) {
+		debug_log(LOG_TYPE_BATMAN, "Drop packet: not received via bidirectional link\n");
+		return;
+	}
+
+	if (is_duplicate) {
+		debug_log(LOG_TYPE_BATMAN, "Drop packet: duplicate packet received\n");
+		return;
+	}
+
+	debug_log(LOG_TYPE_BATMAN, "Forwarding packet: rebroadcast originator packet \n");
+	schedule_forward_packet(orig_node, ethhdr, batman_packet, 0, hna_buff_len, if_incoming);
 }
 
 void purge_orig(struct work_struct *work)
@@ -556,10 +614,11 @@ void purge_orig(struct work_struct *work)
 	}
 
 	spin_unlock(&orig_hash_lock);
+
 	start_purge_timer();
 }
 
-int receive_raw_packet(struct socket *raw_sock, unsigned char *packet_buff, int packet_buff_len)
+static int receive_raw_packet(struct socket *raw_sock, unsigned char *packet_buff, int packet_buff_len)
 {
 	struct kvec iov;
 	struct msghdr msg;
@@ -599,13 +658,14 @@ int packet_recv_thread(void *data)
 
 	while ((!kthread_should_stop()) && (!atomic_read(&exit_cond))) {
 
-
 		wait_event_interruptible(thread_wait, (atomic_read(&data_ready_cond) || atomic_read(&exit_cond)));
 
+		atomic_set(&data_ready_cond, 0);
+
 		if (kthread_should_stop() || atomic_read(&exit_cond))
 			break;
 
-		/* we only want to safely traverse the list, hard-interfaces 
+		/* we only want to safely traverse the list, hard-interfaces
 		 * won't be deleted anyway as long as this thread runs. */
 
 		rcu_read_lock();
@@ -616,12 +676,19 @@ int packet_recv_thread(void *data)
 
 			while (1) {
 				if (batman_if->if_active != IF_ACTIVE) {
-					debug_log(LOG_TYPE_NOTICE, "Could not read from deactivated interface %s!\n", batman_if->dev);
+					if (batman_if->if_active != IF_TO_BE_ACTIVATED)
+						debug_log(LOG_TYPE_NOTICE,
+						          "Could not read from deactivated interface %s!\n",
+						          batman_if->dev);
+
+					if (batman_if->raw_sock)
+						receive_raw_packet(batman_if->raw_sock, packet_buff, PACKBUFF_SIZE);
 					result = 0;
 					break;
 				}
 
-				if ((result = receive_raw_packet(batman_if->raw_sock, packet_buff, PACKBUFF_SIZE)) <= 0)
+				result = receive_raw_packet(batman_if->raw_sock, packet_buff, PACKBUFF_SIZE);
+				if (result <= 0)
 					break;
 
 				if (result < sizeof(struct ethhdr) + 2)
@@ -635,10 +702,9 @@ int packet_recv_thread(void *data)
 					continue;
 				}
 
-				/* batman packet */
 				switch (batman_packet->packet_type) {
+				/* batman originator packet */
 				case BAT_PACKET:
-
 					/* packet with broadcast indication but unicast recipient */
 					if (!is_bcast(ethhdr->h_dest))
 						continue;
@@ -647,15 +713,15 @@ int packet_recv_thread(void *data)
 					if (is_bcast(ethhdr->h_source))
 						continue;
 
-					/* drop packet if it has no batman packet payload */
+					/* drop packet if it has not at least one batman packet as payload */
 					if (result < sizeof(struct ethhdr) + sizeof(struct batman_packet))
 						continue;
 
-					/* network to host order for our 16bit seqno. */
-					batman_packet->seqno = ntohs(batman_packet->seqno);
-
 					spin_lock(&orig_hash_lock);
-					receive_bat_packet(ethhdr, batman_packet, packet_buff + sizeof(struct ethhdr) + sizeof(struct batman_packet), result - sizeof(struct ethhdr) - sizeof(struct batman_packet), batman_if);
+					receive_aggr_bat_packet(ethhdr,
+					                        packet_buff + sizeof(struct ethhdr),
+					                        result - sizeof(struct ethhdr),
+					                        batman_if);
 					spin_unlock(&orig_hash_lock);
 
 					break;
@@ -674,7 +740,7 @@ int packet_recv_thread(void *data)
 					if (!is_my_mac(ethhdr->h_dest))
 						continue;
 
-					/* drop packet if it has not neccessary minimum size */
+					/* drop packet if it has not necessary minimum size */
 					if (result < sizeof(struct ethhdr) + sizeof(struct icmp_packet))
 						continue;
 
@@ -683,92 +749,92 @@ int packet_recv_thread(void *data)
 					/* packet for me */
 					if (is_my_mac(icmp_packet->dst)) {
 
-						/* answer ping request (ping) */
-						if (icmp_packet->msg_type == ECHO_REQUEST) {
-
-							/* get routing information */
-							spin_lock(&orig_hash_lock);
-							orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->orig));
-
-							if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
-
-								memcpy(icmp_packet->dst, icmp_packet->orig, ETH_ALEN);
-								memcpy(icmp_packet->orig, ethhdr->h_dest, ETH_ALEN);
-								icmp_packet->msg_type = ECHO_REPLY;
-								icmp_packet->ttl = TTL;
-
-								send_raw_packet(packet_buff + sizeof(struct ethhdr), result - sizeof(struct ethhdr), orig_node->batman_if->net_dev->dev_addr, orig_node->router->addr, orig_node->batman_if);
-
-							}
-
-							spin_unlock(&orig_hash_lock);
-
-						} else {
-
-							/* add data to device queue */
+						/* add data to device queue */
+						if (icmp_packet->msg_type != ECHO_REQUEST) {
 							bat_device_receive_packet(icmp_packet);
-
-						}
-
-					/* route it */
-					} else {
-
-						/* TTL exceeded */
-						if (icmp_packet->ttl < 2) {
-
-							addr_to_string(src_str, icmp_packet->orig);
-							addr_to_string(dst_str, icmp_packet->dst);
-
-							debug_log(LOG_TYPE_NOTICE, "Error - can't send packet from %s to %s: ttl exceeded\n", src_str, dst_str);
-
-							/* send TTL exceed if packet is an echo request (traceroute) */
-							if (icmp_packet->msg_type == ECHO_REQUEST) {
-
-								/* get routing information */
-								spin_lock(&orig_hash_lock);
-								orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->orig));
-
-								if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
-
-									memcpy(icmp_packet->dst, icmp_packet->orig, ETH_ALEN);
-									memcpy(icmp_packet->orig, ethhdr->h_dest, ETH_ALEN);
-									icmp_packet->msg_type = TTL_EXCEEDED;
-									icmp_packet->ttl = TTL;
-
-									send_raw_packet(packet_buff + sizeof(struct ethhdr), result - sizeof(struct ethhdr), orig_node->batman_if->net_dev->dev_addr, orig_node->router->addr, orig_node->batman_if);
-
-								}
-
-								spin_unlock(&orig_hash_lock);
-
-							}
-
 							continue;
-
 						}
 
+						/* answer echo request (ping) */
 						/* get routing information */
 						spin_lock(&orig_hash_lock);
-						orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->dst));
+						orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->orig));
 
 						if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
 
-							/* decrement ttl */
-							icmp_packet->ttl--;
+							memcpy(icmp_packet->dst, icmp_packet->orig, ETH_ALEN);
+							memcpy(icmp_packet->orig, ethhdr->h_dest, ETH_ALEN);
+							icmp_packet->msg_type = ECHO_REPLY;
+							icmp_packet->ttl = TTL;
 
-							send_raw_packet(packet_buff + sizeof(struct ethhdr), result - sizeof(struct ethhdr), orig_node->batman_if->net_dev->dev_addr, orig_node->router->addr, orig_node->batman_if);
+							send_raw_packet(packet_buff + sizeof(struct ethhdr),
+							                result - sizeof(struct ethhdr),
+							                orig_node->batman_if,
+							                orig_node->router->addr);
 
 						}
 
 						spin_unlock(&orig_hash_lock);
+						continue;
 
 					}
 
+					/* TTL exceeded */
+					if (icmp_packet->ttl < 2) {
+
+						addr_to_string(src_str, icmp_packet->orig);
+						addr_to_string(dst_str, icmp_packet->dst);
+
+						debug_log(LOG_TYPE_NOTICE, "Error - can't send packet from %s to %s: ttl exceeded\n", src_str, dst_str);
+
+						/* send TTL exceeded if packet is an echo request (traceroute) */
+						if (icmp_packet->msg_type != ECHO_REQUEST)
+							continue;
+
+						/* get routing information */
+						spin_lock(&orig_hash_lock);
+						orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->orig));
+
+						if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
+
+							memcpy(icmp_packet->dst, icmp_packet->orig, ETH_ALEN);
+							memcpy(icmp_packet->orig, ethhdr->h_dest, ETH_ALEN);
+							icmp_packet->msg_type = TTL_EXCEEDED;
+							icmp_packet->ttl = TTL;
+
+							send_raw_packet(packet_buff + sizeof(struct ethhdr),
+							                result - sizeof(struct ethhdr),
+							                orig_node->batman_if,
+							                orig_node->router->addr);
+
+						}
+
+						spin_unlock(&orig_hash_lock);
+						continue;
+
+					}
+
+					/* get routing information */
+					spin_lock(&orig_hash_lock);
+					orig_node = ((struct orig_node *)hash_find(orig_hash, icmp_packet->dst));
+
+					if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
+
+						/* decrement ttl */
+						icmp_packet->ttl--;
+
+						/* route it */
+						send_raw_packet(packet_buff + sizeof(struct ethhdr),
+						                result - sizeof(struct ethhdr),
+						                orig_node->batman_if,
+						                orig_node->router->addr);
+					}
+
+					spin_unlock(&orig_hash_lock);
 					break;
 
 				/* unicast packet */
 				case BAT_UNICAST:
-
 					/* packet with unicast indication but broadcast recipient */
 					if (is_bcast(ethhdr->h_dest))
 						continue;
@@ -781,7 +847,7 @@ int packet_recv_thread(void *data)
 					if (!is_my_mac(ethhdr->h_dest))
 						continue;
 
-					/* drop packet if it has not neccessary minimum size - 1 byte ttl, 1 byte payload */
+					/* drop packet if it has not necessary minimum size */
 					if (result < sizeof(struct ethhdr) + sizeof(struct unicast_packet))
 						continue;
 
@@ -791,37 +857,39 @@ int packet_recv_thread(void *data)
 					if (is_my_mac(unicast_packet->dest)) {
 
 						interface_rx(soft_device, packet_buff + sizeof(struct ethhdr) + sizeof(struct unicast_packet), result - sizeof(struct ethhdr) - sizeof(struct unicast_packet));
+						continue;
 
-					/* route it */
-					} else {
-						/* TTL exceeded */
-						if (unicast_packet->ttl < 2) {
-							addr_to_string(src_str, ((struct ethhdr *)(unicast_packet + 1))->h_source);
-							addr_to_string(dst_str, unicast_packet->dest);
-
-							debug_log(LOG_TYPE_NOTICE, "Error - can't send packet from %s to %s: ttl exceeded\n", src_str, dst_str);
-							continue;
-						}
-
-						/* get routing information */
-						spin_lock(&orig_hash_lock);
-						orig_node = ((struct orig_node *)hash_find(orig_hash, unicast_packet->dest));
-
-						if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
-							/* decrement ttl */
-							unicast_packet->ttl--;
-
-							send_raw_packet(packet_buff + sizeof(struct ethhdr), result - sizeof(struct ethhdr), orig_node->batman_if->net_dev->dev_addr, orig_node->router->addr, orig_node->batman_if);
-						}
-
-						spin_unlock(&orig_hash_lock);
 					}
 
+					/* TTL exceeded */
+					if (unicast_packet->ttl < 2) {
+						addr_to_string(src_str, ((struct ethhdr *)(unicast_packet + 1))->h_source);
+						addr_to_string(dst_str, unicast_packet->dest);
+
+						debug_log(LOG_TYPE_NOTICE, "Error - can't send packet from %s to %s: ttl exceeded\n", src_str, dst_str);
+						continue;
+					}
+
+					/* get routing information */
+					spin_lock(&orig_hash_lock);
+					orig_node = ((struct orig_node *)hash_find(orig_hash, unicast_packet->dest));
+
+					if ((orig_node != NULL) && (orig_node->batman_if != NULL) && (orig_node->router != NULL)) {
+						/* decrement ttl */
+						unicast_packet->ttl--;
+
+						/* route it */
+						send_raw_packet(packet_buff + sizeof(struct ethhdr),
+						                result - sizeof(struct ethhdr),
+						                orig_node->batman_if,
+						                orig_node->router->addr);
+					}
+
+					spin_unlock(&orig_hash_lock);
 					break;
 
 				/* broadcast packet */
 				case BAT_BCAST:
-
 					/* packet with broadcast indication but unicast recipient */
 					if (!is_bcast(ethhdr->h_dest))
 						continue;
@@ -830,7 +898,7 @@ int packet_recv_thread(void *data)
 					if (is_bcast(ethhdr->h_source))
 						continue;
 
-					/* drop packet if it has not neccessary minimum size - orig source mac, 2 byte seqno, 1 byte padding, 1 byte payload */
+					/* drop packet if it has not necessary minimum size */
 					if (result < sizeof(struct ethhdr) + sizeof(struct bcast_packet))
 						continue;
 
@@ -847,37 +915,38 @@ int packet_recv_thread(void *data)
 					spin_lock(&orig_hash_lock);
 					orig_node = ((struct orig_node *)hash_find(orig_hash, bcast_packet->orig));
 
-					if (orig_node != NULL) {
-
-						/* check flood history */
-						if (get_bit_status(orig_node->bcast_bits, orig_node->last_bcast_seqno, ntohs(bcast_packet->seqno))) {
-							spin_unlock(&orig_hash_lock);
-							continue;
-						}
-
-						/* mark broadcast in flood history */
-						if (bit_get_packet(orig_node->bcast_bits, ntohs(bcast_packet->seqno) - orig_node->last_bcast_seqno, 1))
-							orig_node->last_bcast_seqno = ntohs(bcast_packet->seqno);
-
-						spin_unlock(&orig_hash_lock);
-						/* broadcast for me */
-						interface_rx(soft_device, packet_buff + sizeof(struct ethhdr) + sizeof(struct bcast_packet), result - sizeof(struct ethhdr) - sizeof(struct bcast_packet));
-
-						/* rebroadcast packet */
-						list_for_each_entry_rcu(batman_if, &if_list, list) {
-
-							send_raw_packet(packet_buff + sizeof(struct ethhdr), result - sizeof(struct ethhdr), batman_if->net_dev->dev_addr, broadcastAddr, batman_if);
-						}
-
-					} else {
+					if (orig_node == NULL) {
 						spin_unlock(&orig_hash_lock);
+						continue;
 					}
 
+					/* check flood history */
+					if (get_bit_status(orig_node->bcast_bits, orig_node->last_bcast_seqno, ntohs(bcast_packet->seqno))) {
+						spin_unlock(&orig_hash_lock);
+						continue;
+					}
+
+					/* mark broadcast in flood history */
+					if (bit_get_packet(orig_node->bcast_bits, ntohs(bcast_packet->seqno) - orig_node->last_bcast_seqno, 1))
+						orig_node->last_bcast_seqno = ntohs(bcast_packet->seqno);
+
+					spin_unlock(&orig_hash_lock);
+
+					/* broadcast for me */
+					interface_rx(soft_device, packet_buff + sizeof(struct ethhdr) + sizeof(struct bcast_packet), result - sizeof(struct ethhdr) - sizeof(struct bcast_packet));
+
+					/* rebroadcast packet */
+					add_bcast_packet_to_list(packet_buff + sizeof(struct ethhdr),
+									result - sizeof(struct ethhdr));
+
 					break;
+
+				/* vis packet */
 				case BAT_VIS:
 					/* drop if too short. */
 					if (result < sizeof(struct ethhdr) + sizeof(struct vis_packet))
 						continue;
+
 					/* not for me */
 					if (!is_my_mac(ethhdr->h_dest))
 						continue;
@@ -888,15 +957,15 @@ int packet_recv_thread(void *data)
 					/* ignore own packets */
 					if (is_my_mac(vis_packet->vis_orig))
 						continue;
+
 					if (is_my_mac(vis_packet->sender_orig))
 						continue;
 
-
 					switch (vis_packet->vis_type) {
 					case VIS_TYPE_SERVER_SYNC:
 						receive_server_sync_packet(vis_packet, vis_info_len);
 						break;
-	
+
 					case VIS_TYPE_CLIENT_UPDATE:
 						receive_client_update_packet(vis_packet, vis_info_len);
 						break;
@@ -904,8 +973,8 @@ int packet_recv_thread(void *data)
 					default:	/* ignore unknown packet */
 						break;
 					}
-					break;
 
+					break;
 				}
 
 			}
@@ -918,8 +987,6 @@ int packet_recv_thread(void *data)
 		}
 		rcu_read_unlock();
 
-		atomic_set(&data_ready_cond, 0);
-
 	}
 	kfree(packet_buff);
 
